/*
 * Program showing a graph that displays values received from a serial connection
 * The data received should be comma separated strings representing integer values
 * In this program there are 4 parameters, adapt if needed
 * For information on the used graph library, check:
 * http://www.gwoptics.org/processing/gwoptics_p5lib/
 *
 * Author P.Peters
 * May 2015
 * Version 0.5
 */

// use serial communication
import processing.serial.*;
// and the gwoptics graph library
import org.gwoptics.graphics.graph2D.Graph2D;
import org.gwoptics.graphics.graph2D.traces.ILine2DEquation;
import org.gwoptics.graphics.graph2D.traces.RollingLine2DTrace;

// character indicating end-of-line
byte linefeed = 10;

// dynamic value indicating which data values are displayed at the top of the graph
int dataMax = 100;

// data array to store the values received from the serial connection after conversion
// length depends on number of parameters received over the serial connection
int[] data = { 0,0,0,0 };

// the number of lines displayed in the graph should be the same as the amount of parameters received
// number of variables depends on number of parameters received over the serial connection
RollingLine2DTrace r0, r1, r2, r3;

// Graph variable
Graph2D g;

// and the serial connection
Serial myPort;

// function (per parameter received) performing possible calculations on the data
// before it is used in line display
// number of these functions depends on number lines displayed in the graph
// (and thus on number of parameters received over the serial connection)
// in this program they all return the actual value without changes
class eq0 implements ILine2DEquation {
  public double computePoint(double x, int pos) {
    return data[0];
  }
}

class eq1 implements ILine2DEquation {
  public double computePoint(double x, int pos) {
    return data[1];
  }
}

class eq2 implements ILine2DEquation {
  public double computePoint(double x, int pos) {
    return data[2];
  }
}

class eq3 implements ILine2DEquation {
  public double computePoint(double x, int pos) {
    return data[3];
  }
}

void setup() {
 // canvas size 
  size(600, 300);

  // create lines to be displayed and define refreshrate, x-increment and their colors
  // number of these depends on number of lines declared before...
  // line 0
  r0  = new RollingLine2DTrace(new eq0(), 200, 0.1f);
  r0.setTraceColour(255, 0, 0);
  // line 1
  r1 = new RollingLine2DTrace(new eq1(), 200, 0.1f);
  r1.setTraceColour(255, 0, 255);
  // line 2
  r2 = new RollingLine2DTrace(new eq2(), 200, 0.1f);
  r2.setTraceColour(0, 0, 255);
  // line 3
  r3 = new RollingLine2DTrace(new eq3(), 200, 0.1f);
  r3.setTraceColour(0, 0, 0);

  // create graph and set graph parameters
  g = new Graph2D(this, 400, 200, false);
  g.setYAxisMax(dataMax);
  g.setYAxisMin(0);
  // add lines to be displayed
  // number of additions depends on number of lines declared before...
  g.addTrace(r0);
  g.addTrace(r1);
  g.addTrace(r2);
  g.addTrace(r3);
  // define position in canvas
  g.position.y = 50;
  g.position.x = 100;
  // default values for Y axis ticks distance
  g.setYAxisTickSpacing(250);
  g.setXAxisMax(5f);
  
  // display list of available serial ports
  for (int i=0; i<Serial.list().length;i++) println("index "+i+" : "+Serial.list()[i]);
  
  // the index used in Serial.list()[index] should be determined
  // when it is known which serial connection to use, this differs
  // per computer depending on how many serial ports you have...
  // choose from what is displayed at first run, it should match the port used
  // in the Arduino IDE (if you are using an Arduino)
  int index = 1;
  // create serial connection
  myPort = new Serial(this, Serial.list()[index], 115200);
  
  // this is used to notify the other side (the Arduino) that we're ready
  // I use it to start sending out the sensor parameters repetitively
  myPort.write("r\n");
} 

void draw() {
  // check data availability
  if (checkForNewData()) {
    // and display it if it is there
    background(255);
    // adapt Y-axis maximum value and tick distance if necessary according to data received
    g.setYAxisMax(dataMax);
    g.setYAxisTickSpacing(dataMax/4);
    // show graph
    g.draw();
  }
}

boolean checkForNewData() {
  // if data available
  if (myPort.available() > 0) { 
    // you know you received a string with comma separated paramaters, so read it
    String inString = myPort.readStringUntil(linefeed);
    // if it is not empty
    if (inString != null) {
      // dislay it (just to check)
      print(inString);
      // and split the string into an array of ints
      // first cleanup the received string (remove leading and trailing spaces and newline)
      // then split into array of strings (using the comma separator)
      // then convert into array of ints
      data = int(split(trim(inString), ',')); 
      // create a new maximum if needed
      for (int i=0; i < data.length; i++) {
        dataMax = max(dataMax,data[i]);
      }
      return true;
    }
  }
  return false;
}

// this is only useful if you actually also check for data reception in the Arduino
// and do something with that data...
void keyPressed() {
  // send the key pressed
  myPort.write(key);
  // and a newline
  myPort.write('\n');
}
